"""Checks the fast Hadamard transform based convolution
feature generation routines to ensure they are producing
correct results by comparing with a clunky, simple
mostly Python implementation."""
import sys

import unittest
import numpy as np
import cupy as cp

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuConv1dFGen, cpuConvGrad, cpuConv1dMaxpool
from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaConv1dFGen, cudaConvGrad, cudaConv1dMaxpool

from conv_testing_functions import get_initial_matrices_fht, get_features
from conv_testing_functions import get_features_with_gradient



class TestConv1d(unittest.TestCase):
    """Tests FHT-based convolution feature generation."""

    def test_conv1d(self):
        """Tests the FHT-based Conv1d C / Cuda functions."""
        # Order: kernel_width, num_aas, aa_dim, num_freqs, sigma, ndatapoints
        test_settings = [
                        (9,23,21,1000,0.5,124),
                        (5,56,2,62,0.5,236),
                        (9,23,21,1000,1,124),
                        (7,202,105,784,1,38),
                        (9,10,2000,4096,1,2),
                        (10,11,200,784,1,38),
                        (10,256,512,333,1,6)
                        ]

        for test_setting in test_settings:
            kernel_width, num_aas, aa_dim, num_freqs, sigma, ndatapoints = test_setting
            outcomes = run_basic_eval(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma)
            for outcome in outcomes:
                self.assertTrue(outcome)
            outcomes = run_basic_eval(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, precision="float")
            for outcome in outcomes:
                self.assertTrue(outcome)




    def test_conv1d_gradients(self):
        """Tests the gradient calculations for the FHT-based Cuda / C functions."""
        kernel_width, num_aas, aa_dim, num_freqs = 9, 23, 21, 128
        sigma, ndatapoints = 0.5, 36

        outcomes = run_gradient_eval(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma)
        for outcome in outcomes:
            self.assertTrue(outcome)

        outcomes = run_gradient_eval(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, precision = "float")
        for outcome in outcomes:
            self.assertTrue(outcome)


        kernel_width, num_aas, aa_dim, num_freqs = 5, 56, 2, 62
        sigma, ndatapoints = 0.5, 53

        outcomes = run_gradient_eval(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma)
        for outcome in outcomes:
            self.assertTrue(outcome)

        outcomes = run_gradient_eval(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, precision = "float")
        for outcome in outcomes:
            self.assertTrue(outcome)

        kernel_width, num_aas, aa_dim, num_freqs = 10, 56, 2000, 4096
        sigma, ndatapoints = 1, 2

        outcomes = run_gradient_eval(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma)
        for outcome in outcomes:
            self.assertTrue(outcome)

        outcomes = run_gradient_eval(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, precision = "float")
        for outcome in outcomes:
            self.assertTrue(outcome)



    def test_normalization(self):
        """Tests that sqrt and full normalization are behaving as
        expected."""
        kernel_width, num_aas, aa_dim, num_freqs = 7, 53, 105, 784
        sigma, ndatapoints = 1, 38

        outcomes = run_basic_eval(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, normalization = 1)
        for outcome in outcomes:
            self.assertTrue(outcome)

        kernel_width, num_aas, aa_dim, num_freqs = 7, 53, 105, 784
        sigma, ndatapoints = 1, 38

        outcomes = run_basic_eval(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, normalization = 2)
        for outcome in outcomes:
            self.assertTrue(outcome)



def run_basic_eval(ndatapoints, kernel_width, aa_dim, num_aas,
        num_freqs, sigma, precision = "double",
        normalization = 0):
    """Run an evaluation of RBF-based convolution kernel feature
    evaluation, without evaluating gradient."""
    dim2, num_blocks, xdata, seqlen, features, s_mat, \
                radem = get_initial_matrices_fht(ndatapoints, kernel_width,
                        aa_dim, num_aas, num_freqs, "conv", precision)
    true_features = get_features(xdata, kernel_width, dim2,
                            radem, s_mat, num_freqs, num_blocks, sigma,
                            seqlen, precision, normalization)
    xd = xdata * sigma
    cpuConv1dFGen(xd, features, radem, s_mat,
            seqlen, kernel_width, normalization, 2)

    outcome = check_results(true_features, features, precision)
    if normalization != 0:
        print(f"****Running test with {normalization} normalization.****")
    print(f"Settings: N {ndatapoints}, kernel_width {kernel_width}, "
        f"aa_dim: {aa_dim}, num_aas: {num_aas}, num_freqs: {num_freqs}, "
        f"sigma: {sigma}, mode: RBF convolution, precision {precision} "
        f"Does result match on CPU? {outcome}")

    if "cupy" not in sys.modules:
        return [outcome]
    xd = cp.asarray(xd)
    features[:] = 0
    features = cp.asarray(features)
    s_mat = cp.asarray(s_mat)
    radem = cp.asarray(radem)
    cudaConv1dFGen(xd, features, radem, s_mat,
            seqlen, kernel_width, normalization)

    features = cp.asnumpy(features)
    outcome_cuda = check_results(true_features, features, precision)
    print(f"Settings: N {ndatapoints}, kernel_width {kernel_width}, "
        f"aa_dim: {aa_dim}, num_aas: {num_aas}, num_freqs: {num_freqs}, "
        f"sigma: {sigma}, mode: RBF convolution, precision {precision} "
        f"Does result match on cuda? {outcome_cuda}")

    print("\n")
    return outcome, outcome_cuda



def run_gradient_eval(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, precision = "double"):
    """Evaluate RBF-based convolution kernel feature evaluation AND
    gradient calculation."""
    dim2, num_blocks, xdata, seqlen, features, s_mat, \
                radem = get_initial_matrices_fht(ndatapoints, kernel_width,
                        aa_dim, num_aas, num_freqs, "conv", precision)
    true_features, true_gradient = get_features_with_gradient(xdata, kernel_width, dim2,
                            radem, s_mat, num_freqs, num_blocks, sigma,
                            seqlen, precision)

    gradient = np.zeros((features.shape[0], features.shape[1], 1))
    cpuConvGrad(xdata, features, radem, s_mat,
            seqlen, gradient, sigma, kernel_width,
            0, 2)
    gradient = gradient[:,:(2*num_freqs),0]

    outcome = check_results(true_features, features[:,:(2 * num_freqs)], precision)
    outcome_gradient = check_results(true_gradient, gradient, precision)
    print(f"Settings: N {ndatapoints}, kernel_width {kernel_width}, "
            f"aa_dim: {aa_dim}, num_aas: {num_aas}, num_freqs: {num_freqs}, "
            f"sigma: {sigma}, mode: RBF conv gradient, precision: {precision} "
            f"Does result match on CPU? {outcome}\n"
            f"Does gradient match on CPU? {outcome_gradient}")

    if "cupy" not in sys.modules:
        return outcome, outcome_gradient

    xdata = cp.asarray(xdata)
    features[:] = 0
    features = cp.asarray(features)
    gradient = cp.zeros((features.shape[0], features.shape[1], 1))
    s_mat = cp.asarray(s_mat)
    radem = cp.asarray(radem)

    cudaConvGrad(xdata, features, radem, s_mat,
            seqlen, gradient, sigma, kernel_width, 0)
    features = cp.asnumpy(features[:,:(2*num_freqs)])
    gradient = cp.asnumpy(gradient[:,:(2*num_freqs),0])

    outcome_cuda = check_results(true_features, features, precision)
    outcome_cuda_gradient = check_results(true_gradient, gradient, precision)
    print(f"Settings: N {ndatapoints}, kernel_width {kernel_width}, "
        f"aa_dim: {aa_dim}, num_aas: {num_aas}, num_freqs: {num_freqs}, "
        f"sigma: {sigma}, mode: RBF conv gradient, precision {precision} "
        f"Does result match on cuda? {outcome_cuda}\n"
        f"Does gradient match on cuda? {outcome_cuda_gradient}\n")

    print("\n")
    return outcome, outcome_gradient, outcome_cuda, outcome_cuda_gradient





def check_results(gt_array, test_array, precision):
    """Checks a ground truth array against a test array. We have
    to use different tolerances for 32-bit vs 64 since 32-bit
    can vary slightly."""
    if precision == "float":
        return np.allclose(gt_array, test_array, rtol=1e-5, atol=1e-5)
    return np.allclose(gt_array, test_array)



if __name__ == "__main__":
    unittest.main()



"""Checks that non-zero-padded variable length sequences can
be input to an FHTConv1d kernel with no errors."""
import sys
import os
import unittest
import numpy as np

try:
    import cupy as cp
except:
    pass

from xGPR import build_regression_dataset
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.model_constructor import get_models


#A list of the kernels to be tested.
variable_length_kernels = ["Conv1dRBF"]


class TestVariableLengthSeqs(unittest.TestCase):
    """Tests for errors when running non-zero-padded variable
    length sequences through a graph or sequence kernel."""

    def test_for_length_errors(self):
        """Tests the Conv1dRBF kernel."""
        rng = np.random.default_rng(123)
        block1 = rng.uniform(size=(100,20,21))
        block2 = rng.uniform(size=(100,10,21))
        dud_block = rng.uniform(size=(100,2,21))
        for kernel in variable_length_kernels:
            outcomes = run_kernel_specific_test(kernel, block1, block2,
                    dud_block)
            for outcome in outcomes:
                self.assertTrue(outcome)


def run_kernel_specific_test(kernel, block1, block2, dud_block):
    """Checks the specified kernel using both cpu and (if available)
    cuda on two valid input blocks (block1 and block2) and an invalid
    input block. Returns True for each test that passed and False
    for each that failed. This is a simple does it work or does it
    raise an exception test -- the correctness of the FHT operations
    is tested under the other tests in this folder."""
    sequence_lengths = np.full(block1.shape[0],
                block1.shape[1]).astype(np.int32)
    online_dataset = build_regression_dataset(block1,
            np.zeros((block1.shape[0])), sequence_lengths)
    models = get_models(kernel, online_dataset,
                        num_rffs = 1024,
                        conv_ard_kernel = False)
    outcomes = []
    for (model, device) in zip(models, ["cpu", "cuda"]):
        if model is None:
            continue

        sequence_lengths = np.full(block1.shape[0],
                block1.shape[1]).astype(np.int32)
        try:
            _ = model.kernel.transform_x(block1, sequence_lengths)
            outcomes.append(True)
        except:
            outcomes.append(False)

        try:
            sequence_lengths = np.full(block2.shape[0],
                    block2.shape[1]).astype(np.int32)
            _ = model.kernel.transform_x(block2, sequence_lengths)
            outcomes.append(True)
        except:
            outcomes.append(False)

        try:
            sequence_lengths = np.full(dud_block.shape[0],
                    dud_block.shape[1]).astype(np.int32)
            _ = model.kernel.transform_x(dud_block, sequence_lengths)
            outcomes.append(False)
        except:
           outcomes.append(True)

        try:
            sequence_lengths = np.full(dud_block.shape[0],
                    dud_block.shape[1]).astype(np.int32)
            _ = model.kernel.transform_x(block2, sequence_lengths)
            outcomes.append(False)
        except:
           outcomes.append(True)

        try:
            sequence_lengths = np.full(dud_block.shape[0], 21).astype(np.int32)
            _ = model.kernel.transform_x(block2, sequence_lengths)
            outcomes.append(False)
        except:
           outcomes.append(True)

    return outcomes



if __name__ == "__main__":
    unittest.main()



"""Contains functions for calculating exact and approximate
NMLL gradients for the regression model."""
import numpy as np
try:
    import cupy as cp
    import cupyx as cpx
except:
    pass
from scipy.linalg import solve_triangular, cho_solve
from .alpha_beta_optimizer import optimize_alpha_beta

def calc_gradient_terms(dataset, kernel, device, subsample = 1):
    """Calculates terms needed for the gradient calculation.
    Specific for negative marginal log likelihood (NMLL),
    as opposed to loss on the training set.

    Args:
        dataset: An OnlineDataset or OfflineDataset with the
            raw data we need for these calculations.
        kernel: A valid kernel object that can generate random features.
        device (str): One of "cpu", "cuda".
        subsample (float): A value in the range [0.01,1] that indicates what
            fraction of the training set to use each time the gradient is
            calculated (the same subset is used every time). In general, 1
            will give better results, but using a subsampled subset can be
            a fast way to find the (approximate) location of a good
            hyperparameter set.

    Returns:
        z_trans_z: The M x M cupy or numpy matrix for M random features.
        z_trans_y: The length M cupy or numpy array for M random features.
        y_trans_y (float): The dot product of y with itself.
        dz_dsigma_ty (array): Derivative w/r/t kernel-specific hyperparams times y.
        inner_deriv (array): Derivative for the log determinant portion of the NMLL.
        ndatapoints (int): The number of datapoints.
    """
    if subsample > 1 or subsample < 0.01:
        raise ValueError("Subsample must be in the range [0.01, 1].")

    num_rffs = kernel.get_num_rffs()
    hparams = kernel.get_hyperparams()
    if device == "cpu":
        z_trans_z = np.zeros((num_rffs, num_rffs))
        z_trans_y = np.zeros((num_rffs))
        dz_dsigma_ty = np.zeros((num_rffs, hparams.shape[0] - 1))
        inner_deriv = np.zeros((num_rffs, num_rffs,
                                hparams.shape[0] - 1))
        transpose = np.transpose
    else:
        z_trans_z = cp.zeros((num_rffs, num_rffs))
        z_trans_y = cp.zeros((num_rffs))
        dz_dsigma_ty = cp.zeros((num_rffs, hparams.shape[0] - 1))
        inner_deriv = cp.zeros((num_rffs, num_rffs,
                                    hparams.shape[0] - 1))
        transpose = cp.transpose

    y_trans_y = 0
    ndatapoints = 0

    if subsample == 1:
        for xin, yin, ldata in dataset.get_chunked_data():
            xfeatures, dz_dsigma, ydata = kernel.gradient_x_y(xin, yin, ldata)
            z_trans_y += xfeatures.T @ ydata
            z_trans_z += xfeatures.T @ xfeatures
            y_trans_y += ydata.T @ ydata
            ndatapoints += xfeatures.shape[0]

            for i in range(dz_dsigma.shape[2]):
                dz_dsigma_ty[:,i] += dz_dsigma[:,:,i].T @ ydata
                inner_deriv[:,:,i] += dz_dsigma[:,:,i].T @ xfeatures
    else:
        rng = np.random.default_rng(123)
        for xin, yin, ldata in dataset.get_chunked_data():
            idx_size = max(1, int(subsample * xin.shape[0]))
            idx = rng.choice(xin.shape[0], idx_size, replace=False)
            xin, yin, ldata = xin[idx,...], yin[idx], ldata[idx]
            xfeatures, dz_dsigma, ydata = kernel.gradient_x_y(xin, yin, ldata)

            z_trans_y += xfeatures.T @ ydata
            z_trans_z += xfeatures.T @ xfeatures
            y_trans_y += ydata.T @ ydata
            ndatapoints += xin.shape[0]

            for i in range(dz_dsigma.shape[2]):
                dz_dsigma_ty[:,i] += dz_dsigma[:,:,i].T @ ydata
                inner_deriv[:,:,i] += dz_dsigma[:,:,i].T @ xfeatures

    inner_deriv += transpose(inner_deriv, (1,0,2))

    return z_trans_z, z_trans_y, float(y_trans_y), dz_dsigma_ty, inner_deriv, ndatapoints



def exact_nmll_reg_grad(z_trans_z, z_trans_y, y_trans_y,
                                hparams, ndatapoints,
                                dz_dsigma_ty, inner_deriv, device):
    """Calculates the gradient for NMLL. The gradient calculation here is
    for regression.

    Args:
        z_trans_z_chol: Numpy or cupy array of shape (num_rffs, num_rffs)
            containing z^T z.
        z_trans_y: Numpy or cupy array of shape (num_rffs) containing
            z^T y.
        y_trans_y (float): The dot product y^T y.
        hparams (np.ndarray): The hyperparameters (not the log of the hyperparameters).
        beta (float): The selected value for the nuisance parameter beta.
        dz_dsigma_ty: A cupy or numpy array containing (dz_dsigma^T y).
            Shape is (num_rffs, M) where M is the number of kernel-
            specific hyperparameters.
        ndatapoints (int): The number of datapoints.
        device (str): Either "cpu" or "cuda".

    Returns:
        grad (np.ndarray): A numpy array containing the gradient of the
            hyperparameters.
    """
    z_trans_z.flat[::z_trans_z.shape[0]+1] += hparams[0]**2

    if device == "cuda":
        weights, z_trans_z_chol, chol_inv = gpu_cho_calcs(z_trans_z,
                    z_trans_y, hparams[0])
        cho_solver = gpu_cho_solver
        nll1 = float(0.5 * (y_trans_y - z_trans_y.T @ weights))
        nll2 = float(cp.log(cp.diag(z_trans_z_chol)).sum())
    else:
        weights, z_trans_z_chol, chol_inv = cpu_cho_calcs(z_trans_z,
                    z_trans_y, hparams[0])
        cho_solver = cpu_cho_solver
        nll1 = float(0.5 * (y_trans_y - z_trans_y.T @ weights))
        nll2 = float(np.log(np.diag(z_trans_z_chol)).sum())

    nrffs = float(z_trans_z.shape[0])
    negloglik, beta = optimize_alpha_beta(hparams[0],
                np.array([nll1, nll2]), float(ndatapoints), nrffs)

    grad = np.zeros((hparams.shape[0]))

    #Note that in the following, lambda_ is hparams[0], beta is a
    #nuisance parameter, and alpha is lambda_ * beta.
    alpha = hparams[0] * beta

    #First calculate gradient w/r/t lambda...
    dnll_dlambda = (1 / (beta**2 * hparams[0]**3)) * ((z_trans_y.T @ weights) - y_trans_y)
    dnll_dlambda += (1 / (beta**2 * hparams[0])) * (weights.T @ weights)
    dnll_dlambda += (ndatapoints - z_trans_z_chol.shape[1]) / hparams[0]
    dnll_dlambda += hparams[0] * (chol_inv**2).sum()
    grad[0] = float(dnll_dlambda)

    #Finally, calculate kernel-specific hyperparameter gradients.

    for i in range(grad.shape[0] - 1):
        trace_term = cho_solver(z_trans_z_chol, inner_deriv[:,:,i])
        dnll_dsigma = -2 * (weights.T @ dz_dsigma_ty[:,i])
        dnll_dsigma += (weights.T @ (inner_deriv[:,:,i] @ weights))
        dnll_dsigma *= (0.5 / alpha**2)
        dnll_dsigma += 0.5 * trace_term.trace()
        grad[i+1] = float(dnll_dsigma)

    grad *= hparams
    return negloglik, grad, beta


def cpu_cho_solver(chol_decomp, target):
    """Provides a cho_solver calc for cpu (so that
    both the cpu and gpu call share a common interface)."""
    return cho_solve((chol_decomp, True), target)

def gpu_cho_solver(chol_decomp, target):
    """Provides a cho_solver calc for gpu (so that
    both the cpu and gpu call share a common interface).
    TODO: Optimize."""
    sol = cpx.scipy.linalg.solve_triangular(chol_decomp,
                            target, lower=True)
    return cpx.scipy.linalg.solve_triangular(chol_decomp.T,
                           sol, lower=False)


def gpu_cho_calcs(z_trans_z, z_trans_y, lambda_):
    """Calculates some key quantities (e.g. the inner derivative)
    needed for shared gradient calculations. TODO: This can be
    made more efficient, and should be rewritten / optimized
    at a lower level.

    Args:
        z_trans_z: cupy array of shape (self.num_rffs, self.num_rffs)
            containing z^T z.
        z_trans_y: cupy array of shape (self.num_rffs) containing z^T y.
        lambda_ (float): The value of the first hyperparameter (noise) shared
            between all kernels.

    Returns:
        weights: cupy array of shape (self.num_rffs) containing the
            weights.
        z_trans_z_chol: A cupy array storing the cholesky decomposition
            of z^T z. Shape is (self.num_rffs, self.num_rffs).
        id_trace (float): The trace of (Z^T Z + lambda**2)^-1 (Z^T Z).
        chol_inv: A cupy array of shape (self.num_rffs, self.num_rffs)
            storing the inverse of z_trans_z_chol.
    """
    z_trans_z_chol = cp.linalg.cholesky(z_trans_z)
    weights = gpu_cho_solver(z_trans_z_chol, z_trans_y)
    z_trans_z.flat[::z_trans_z.shape[0]+1] -= lambda_**2
    chol_inv = cpx.scipy.linalg.solve_triangular(z_trans_z_chol,
                    cp.eye(z_trans_z_chol.shape[0]), lower=True)
    return weights, z_trans_z_chol, chol_inv


def cpu_cho_calcs(z_trans_z, z_trans_y, lambda_):
    """Calculates some key quantities (e.g. the inner derivative)
    needed for shared gradient calculations. TODO: This can be
    made more efficient, and should be rewritten / optimized
    at a lower level.

    Args:
        z_trans_z: Numpy array of shape (self.num_rffs, self.num_rffs)
            containing z^T z.
        z_trans_y: Numpy array of shape (self.num_rffs) containing z^T y.
        lambda_ (float): The value of the first hyperparameter (noise) shared
            between all kernels.

    Returns:
        weights: numpy array of shape (self.num_rffs) containing the
            weights.
        z_trans_z_chol: A numpy array storing the cholesky decomposition
            of z^T z. Shape is (self.num_rffs, self.num_rffs).
        id_trace (float): The trace of (Z^T Z + lambda**2)^-1 (Z^T Z).
        chol_inv: A numpy array of shape (self.num_rffs, self.num_rffs)
            storing the inverse of z_trans_z_chol.
    """
    z_trans_z_chol = np.linalg.cholesky(z_trans_z)
    weights = cho_solve((z_trans_z_chol, True), z_trans_y)
    z_trans_z.flat[::z_trans_z.shape[0]+1] -= lambda_**2
    chol_inv = solve_triangular(z_trans_z_chol,
                    np.eye(z_trans_z_chol.shape[0]), lower=True)
    return weights, z_trans_z_chol, chol_inv



"""A 'mini-ARD' kernel that assigns one lengthscale per
feature category for different feature types.
Accepts only fixed vectors as input."""
from math import ceil

import numpy as np
from scipy.stats import chi

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuFastHadamardTransform2D as dFHT2d
from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuRBFFeatureGen, cpuMiniARDGrad
try:
    import cupy as cp
    from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaRBFFeatureGen, cudaMiniARDGrad
except:
    pass
from ..kernel_baseclass import KernelBaseclass


class MiniARD(KernelBaseclass):
    """The MiniARD is a subset of automatic relevance determination
    (ARD), which assigns a lengthscale to every feature. MiniARD
    rather assigns a different lengthscale to each group of features
    (e.g. the first 20, the next 20 and so on), as specified by
    a user-provided list of "split points", points at which to split
    up the feature vector.

    This class inherits from KernelBaseclass. Only attributes unique
    to this child are described in this docstring.

    Attributes:
        hyperparams (np.ndarray): A length one + number of feature
            classes array of hyperparameters: lambda_ (noise)
            and the remainder are inverse lengthscales.
        cosfunc: A convenience reference to either cp.cos or np.cos,
            as appropriate for current device.
        sinfunc: A convenience reference to either cp.sin or np.sin,
            as appropriate for current device.
        split_pts (list): The points at which to split the
            input data; a list of up to length 3. I.e.,
            if x is the input data, lengthscale 1 applies
            to x[:,0:split_pts[0]], lengthscale 2 applies
            to x[:,split_pts[0]:split_pts[1]], etc.
        full_ard_weights (array): A cupy or numpy array (depending on device)
            containing the ARD weights for each input feature. This is
            repopulated whenever the hyperparameters are updated.
    """

    def __init__(self, xdim, num_rffs, random_seed = 123,
                device = "cpu", num_threads = 2, double_precision = False,
                kernel_spec_parms = {}):
        """Constructor.

        Args:
            xdim (tuple): The dimensions of the input.
            num_rffs (int): The user-requested number of random Fourier features.
            random_seed (int): The seed to the random number generator.
            device (str): One of 'cpu', 'cuda'. Indicates the starting device.
            num_threads (int): The number of threads to use for generating random
                features if running on CPU. Ignored if running on GPU.
            double_precision (bool): Whether to use double precision for the FHT
                when generating random features. It is generally best to set to
                False -- setting to True increases computational cost for negligible
                benefit.
            kernel_spec_parms (dict): A dictionary of kernel-specific parameters.
                In this case, must contain "split_points", which is a list
                of split points across input arrays.

        Raises:
            ValueError: A ValueError is raised if the needed kernel_spec_parms
                are not supplied, or if the split points supplied are invalid.
        """
        super().__init__(num_rffs, xdim, num_threads, sine_cosine_kernel = True,
                double_precision = double_precision,
                kernel_spec_parms = kernel_spec_parms)
        if len(self._xdim) != 2:
            raise ValueError("The dimensionality of the input is inappropriate for "
                        "the kernel you have selected.")
        if "split_points" not in kernel_spec_parms:
            raise ValueError("For the MiniARD kernel, 'kernel_specific_params' "
                    "must contain a list called 'split_points'.")
        if not isinstance(kernel_spec_parms["split_points"], list):
            raise ValueError("For the MiniARD kernel, 'split_points' must "
                    "be a list.")
        self.split_pts = np.sort([0] + kernel_spec_parms["split_points"] + [xdim[1]])
        self.check_split_points(xdim)

        self.hyperparams = np.ones((self.split_pts.shape[0]))
        bounds = [[1e-3,1e2]] + [[1e-6, 1e2] for i in
                range(self.hyperparams.shape[0] - 1)]
        self.bounds = np.asarray(bounds)

        self.padded_dims = 2**ceil(np.log2(max(xdim[-1], 2)))

        radem_array = np.asarray([-1,1], dtype=np.int8)
        rng = np.random.default_rng(random_seed)
        if self.padded_dims < self.num_freqs:
            self.nblocks = ceil(self.num_freqs / self.padded_dims)
        else:
            self.nblocks = 1
        self.radem_diag = rng.choice(radem_array, size=(3, 1, self.nblocks * self.padded_dims),
                                replace=True)
        self.chi_arr = chi.rvs(df=self.padded_dims, size=self.num_freqs,
                            random_state = random_seed)


        #Converts the hyperparameters into a "full" array of the
        #same length as padded dims, just as if this were an ARD.
        #Also, store the positions that map to different hparams.
        self.full_ard_weights = np.zeros((xdim[-1]))
        self.ard_position_key = np.zeros((xdim[-1]), dtype=np.int32)

        if not self.double_precision:
            self.chi_arr = self.chi_arr.astype(np.float32)

        self.kernel_specific_set_hyperparams()

        self.precomputed_weights = None
        self.device = device



    def check_split_points(self, xdim):
        """Called during initialization to ensure the split points supplied
        by user are valid, or at any time split points are updated.

        Args:
            xdim (tuple): The shape of the anticipated input data.

        Raises:
            ValueError: A ValueError is raised if the splits supplied
                by user are not valid.
        """
        n_splits = self.split_pts.shape[0] - 2
        if n_splits < 1:
            raise ValueError("There must be at least one split point to use MiniARD.")
        if self.split_pts[0] < 0:
            raise ValueError("The first split point must be > 0.")
        if self.split_pts[-1] > xdim[1]:
            raise ValueError("The last split point must be < shape[1] of the "
                    "input data.")
        if np.diff(self.split_pts).min() == 0:
            raise ValueError("At least two of the split points supplied are "
                        "identical.")


    def kernel_specific_set_device(self, new_device):
        """Called by parent class when device is changed. Moves
        some of the object parameters to the appropriate device
        and updates self.cosfunc, self.sinfunc, which are
        convenience references to np.cos / np.sin or cp.cos
        / cp.sin."""
        if new_device == "cpu":
            if not isinstance(self.radem_diag, np.ndarray):
                if self.precomputed_weights is not None:
                    self.precomputed_weights = cp.asnumpy(self.precomputed_weights)
                self.radem_diag = cp.asnumpy(self.radem_diag)
                self.full_ard_weights = cp.asnumpy(self.full_ard_weights)
                self.chi_arr = cp.asnumpy(self.chi_arr)
                self.ard_position_key = cp.asnumpy(self.ard_position_key)
                self.chi_arr = cp.asnumpy(self.chi_arr)
        else:
            self.radem_diag = cp.asarray(self.radem_diag)
            self.chi_arr = cp.asarray(self.chi_arr)
            if self.precomputed_weights is not None:
                self.precomputed_weights = cp.asarray(self.precomputed_weights)
            self.full_ard_weights = cp.asarray(self.full_ard_weights)
            self.ard_position_key = cp.asarray(self.ard_position_key)


    def kernel_specific_set_hyperparams(self):
        """Once hyperparameters have been reset, this kernel needs
        to repopulate an array it will use when generating random
        features."""
        self.full_ard_weights[:] = 0
        self.ard_position_key[:] = 0
        for i in range(1, self.split_pts.shape[0]):
            self.full_ard_weights[self.split_pts[i-1]:self.split_pts[i]] = \
                    self.hyperparams[i]
            self.ard_position_key[self.split_pts[i-1]:self.split_pts[i]] = i - 1


    def kernel_specific_transform(self, input_x, sequence_length = None):
        """Generates random features for an input array.

        Args:
            x: Either a cupy or numpy array containing the input.
            sequence_length: Accepted for consistency with baseclass
                but not used by this kernel and thus ignored.

        Returns:
            xtrans: A cupy or numpy array containing the generated features.
        """
        xtrans = input_x * self.full_ard_weights[None,:]

        if self.device == "cpu":
            output_x = np.zeros((input_x.shape[0], self.num_rffs), np.float64)
            if not self.double_precision:
                xtrans = xtrans.astype(np.float32)
            cpuRBFFeatureGen(xtrans, output_x, self.radem_diag, self.chi_arr,
                    self.num_threads, self.fit_intercept)
        else:
            output_x = cp.zeros((input_x.shape[0], self.num_rffs), cp.float64)
            if not self.double_precision:
                xtrans = xtrans.astype(cp.float32)
            cudaRBFFeatureGen(xtrans, output_x, self.radem_diag, self.chi_arr,
                    self.fit_intercept)
        return output_x


    def precompute_weights(self):
        """The kernel does not automatically generate precomputed weights,
        because for generating features during fitting or prediction,
        these are not necessary and would increase model size unnecessarily.
        Only during gradient calculations are they required. The first time
        the kernel is asked to generate a gradient, then, it will precompute
        and store weights.

        Note that precomputing weights is only helpful for ARD kernels
        because of the greater complexity of calculating the gradient
        using FHT only."""
        precomp_weights = []
        #Currently the FHT-2d only operation is only implemented for
        #CPU. Since the precompute weights operation is only performed
        #once, it has not been a high priority for optimization.
        #TODO: Add this for GPU, and transfer the operations under
        #this function to C / Cuda code.
        if self.device == "cuda":
            self.radem_diag = cp.asnumpy(self.radem_diag)
            self.chi_arr = cp.asnumpy(self.chi_arr)

        norm_constant = np.log2(self.padded_dims) / 2.0
        norm_constant = 1.0 / (2.0**norm_constant)

        padded_chi_arr = np.zeros((self.nblocks * self.padded_dims))
        padded_chi_arr[:self.chi_arr.shape[0]] = self.chi_arr

        for i in range(self.nblocks):
            ident_mat = np.eye(self.padded_dims)
            init_pt, cut_pt = i * self.padded_dims, (i+1) * self.padded_dims
            ident_mat *= self.radem_diag[0:1,0,init_pt:cut_pt] * norm_constant
            dFHT2d(ident_mat, self.num_threads)
            ident_mat *= self.radem_diag[1:2,0,init_pt:cut_pt] * norm_constant
            dFHT2d(ident_mat, self.num_threads)
            ident_mat *= self.radem_diag[2:3,0,init_pt:cut_pt] * norm_constant
            dFHT2d(ident_mat, self.num_threads)


            ident_mat *= padded_chi_arr[init_pt:cut_pt]
            precomp_weights.append(ident_mat.T[:,:self._xdim[-1]])


        self.precomputed_weights = np.vstack(precomp_weights)[:self.num_freqs,:]
        if not self.double_precision:
            self.precomputed_weights = self.precomputed_weights.astype(np.float32)

        self.precomputed_weights = np.ascontiguousarray(self.precomputed_weights)
        if self.device == "cuda":
            self.radem_diag = cp.asarray(self.radem_diag)
            self.precomputed_weights = cp.asarray(self.precomputed_weights)
            self.chi_arr = cp.asarray(self.chi_arr)


    def kernel_specific_gradient(self, input_x, sequence_length = None):
        """The gradient for kernel-specific hyperparameters is calculated
        using an array (dz_dsigma) specific to each kernel.

        Args:
            input_x: A cupy or numpy array containing the raw input data.
            sequence_length: Accepted for consistency with baseclass
                but not used by this kernel and thus ignored.

        Returns:
            output_x: A cupy or numpy array containing the random feature
                representation of the input.
            dz_dsigma: A cupy or numpy array containing the derivative of
                output_x with respect to the kernel-specific hyperparameters.
        """
        if self.precomputed_weights is None:
            self.precompute_weights()
        max_map_position = int(self.ard_position_key.max())

        if self.device == "cpu":
            xtrans = np.zeros((input_x.shape[0], self.num_rffs), np.float64)
            dz_dsigma = np.zeros((input_x.shape[0], self.num_rffs,
                max_map_position + 1), np.float64)
            cpuMiniARDGrad(input_x, xtrans, self.precomputed_weights,
                self.ard_position_key, self.full_ard_weights,
                dz_dsigma, self.num_threads, self.fit_intercept)
        else:
            xtrans = cp.zeros((input_x.shape[0], self.num_rffs), cp.float64)
            dz_dsigma = cp.zeros((input_x.shape[0], self.num_rffs,
                max_map_position + 1), cp.float64)
            cudaMiniARDGrad(input_x, xtrans, self.precomputed_weights,
                self.ard_position_key, self.full_ard_weights,
                dz_dsigma, self.fit_intercept)

        return xtrans, dz_dsigma



"""This is the baseclass for OfflineDataset and OnlineDataset.

The DatasetBaseclass class contains methods shared by both
in-memory and on-disk datasets.
"""
import abc
from abc import ABC


class DatasetBaseclass(ABC):
    """DatasetBaseclass stores methods common to both
    OfflineDataset and OnlineDataset, and ensures
    that both share a common API.

    Attributes:
        xdim_ (tuple): A length 3 tuple (for 3d data) or 2
            (for 2d data) indicating the full dimensions of
            the dataset.
        chunk_size (int): Either the chunk_size for online datasets or the
            largest allowed array size for offline datasets.
        trainy_mean (float): The mean of the training y-data (for regression only).
        trainy_std (float): The standard deviation of the training y-data (for
            regression only).
        max_class (int): The largest class number (for classification only).
    """
    def __init__(self, xdim, chunk_size,
            trainy_mean, trainy_std, max_class=0):
        self._xdim = xdim

        #Trainy_mean and trainy_std are used
        #for regression, while max_class is used
        #for classification.
        self._trainy_mean = trainy_mean
        self._trainy_std = trainy_std
        self._max_class = max_class
        self._chunk_size = chunk_size


    @abc.abstractmethod
    def get_chunked_data(self):
        """Abstract method to force child class to implement
        get_chunked_data."""

    @abc.abstractmethod
    def get_chunked_x_data(self):
        """Abstract method to force child class to implement
        get_chunked_x_data."""


    def get_ymean(self):
        """Returns the mean of the training y data.
        Only useful if data is for regression."""
        return self._trainy_mean

    def get_ystd(self):
        """Returns the standard deviation of the training
        y data. Only useful if data is for regression."""
        return self._trainy_std


    def get_n_classes(self):
        """Gets the largest category number in the training
        data. Only useful if data is for classification."""
        return self._max_class + 1


    def get_xdim(self):
        """Returns the xdim list describing the size of the
        dataset."""
        return self._xdim

    def get_ndatapoints(self):
        """Returns the number of datapoints."""
        return self._xdim[0]

    def get_chunk_size(self):
        """Return the chunk size."""
        return self._chunk_size



"""Contains functions for calculating exact and approximate
NMLL gradients for the regression model."""
import numpy as np
from scipy.linalg import eigh_tridiagonal
try:
    import cupy as cp
except:
    pass



def estimate_logdet(alphas, betas, num_rffs, preconditioner = None,
        device = "cpu"):
    """Estimates the log determinant term in the NMLL using the alpha
    and beta values from a conjugate gradients run against samples
    from a normal distribution with covariance matrix P (if
    preconditioning was used) or I (if no preconditioning).

    Args:
        alphas (np.ndarray): A numpy array of shape (niter, nprobes).
        betas (np.ndarray): A numpy array of shape (niter, nprobes).
        num_rffs (int): The number of RFFs.
        preconditioner: Either None or a valid preconditioner object.

    Returns:
        logdets (float): The estimated log determinant of
            (lambda_ + Z^T Z).
    """
    mat_diag = 1 / alphas
    mat_diag[1:,:] += betas[:-1,:] / alphas[:-1,:]
    upper_diag = np.sqrt(betas) / alphas
    if device == "cuda":
        logdets = cp.zeros((mat_diag.shape[1]))
        tr_estim = cp.zeros((mat_diag.shape[1]))
    else:
        logdets = np.zeros((mat_diag.shape[1]))
        tr_estim = np.zeros((mat_diag.shape[1]))


    for i in range(mat_diag.shape[1]):
        eigvals, eigvecs = eigh_tridiagonal(mat_diag[:,i],
                        upper_diag[:-1,i], lapack_driver = "stev")
        weights = eigvecs[0,:]**2
        logdets[i] += (weights * np.log(eigvals)).sum()
        tr_estim[i] += (weights / eigvals).sum()

    logdets = num_rffs * logdets.sum() / alphas.shape[1]
    if preconditioner is not None:
        logdets += preconditioner.get_logdet()
    return float(logdets)



"""Tests fitting with an offline dataset (to ensure results
are the same)."""
import sys
import os
import unittest

import numpy as np

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset
from utils.model_constructor import get_models
from utils.evaluate_model import evaluate_model

#A set of hyperparameters known to work well for our testing dataset
#that we can use as a default.
HPARAM = np.array([np.log(np.sqrt(0.0767)),  np.log(0.358)])

NUM_RFFS = 4100


class CheckCGFit(unittest.TestCase):
    """Tests the conjugate gradients algorithm."""

    def test_preconditioned_cg(self):
        """Test using preconditioned cg, which should easily fit
        in under 10 epochs."""
        _, offline_data = build_test_dataset(conv_kernel = False)
        cpu_mod, gpu_mod = get_models("RBF", offline_data, num_rffs = NUM_RFFS)

        cpu_mod.set_hyperparams(HPARAM, offline_data)
        preconditioner, _ = cpu_mod.build_preconditioner(offline_data,
            max_rank = 256, method = "srht")

        niter, _ = cpu_mod.fit(offline_data,  preconditioner = preconditioner,
                max_iter = 500, run_diagnostics = True,
                tol = 1e-6,  mode = "cg")
        print(f"niter: {niter}")
        self.assertTrue(niter < 10)

        if gpu_mod is not None:
            gpu_mod.set_hyperparams(HPARAM, offline_data)
            preconditioner, _ = gpu_mod.build_preconditioner(offline_data,
                max_rank = 256, method = "srht")

            niter, _ = gpu_mod.fit(offline_data,  preconditioner = preconditioner,
                max_iter = 500, run_diagnostics = True,
                tol = 1e-6,  mode = "cg")
            print(f"niter: {niter}")
            self.assertTrue(niter < 10)


    def test_autoselect_cg(self):
        """Test using preconditioner autoselect."""
        _, offline_data = build_test_dataset(conv_kernel = False)
        cpu_mod, gpu_mod = get_models("RBF", offline_data)
        cpu_mod.fitting_rffs = NUM_RFFS
        cpu_mod.verbose = False

        cpu_mod.set_hyperparams(HPARAM, offline_data)
        niter, _ = cpu_mod.fit(offline_data,
                max_iter = 500, run_diagnostics = True,
                tol = 1e-6,  mode = "cg")
        print(f"Autoselect preconditioning, niter: {niter}")
        self.assertTrue(niter < 80)

        if gpu_mod is not None:
            gpu_mod.fitting_rffs = NUM_RFFS
            gpu_mod.verbose = False
            gpu_mod.set_hyperparams(HPARAM, offline_data)
            niter, _ = gpu_mod.fit(offline_data,
                max_iter = 500, run_diagnostics = True,
                tol = 1e-6,  mode = "cg")
            print(f"Autoselect preconditioning, niter: {niter}")
            self.assertTrue(niter < 80)


if __name__ == "__main__":
    unittest.main()



"""Tests dataset construction and ensures that y_mean, y_std
are calculated correctly."""
import sys
import os
import unittest

import numpy as np

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset



class CheckDatasetConstruction(unittest.TestCase):
    """Tests construction of dataset objects."""

    def test_dataset_builders(self):
        """Test the dataset builders."""
        test_online_dataset, test_offline_dataset = build_test_dataset(conv_kernel = False,
            xsuffix = "testxvalues.npy", ysuffix = "testyvalues.npy")
        train_online_dataset, train_offline_dataset = build_test_dataset(conv_kernel = False)

        #Access protected class members directly
        test_ymean = np.mean(test_online_dataset._ydata)
        test_ystd = np.std(test_online_dataset._ydata)
        train_ymean = np.mean(train_online_dataset._ydata)
        train_ystd = np.std(train_online_dataset._ydata)

        self.assertTrue(np.allclose(test_ymean, test_offline_dataset.get_ymean()))
        self.assertTrue(np.allclose(test_ystd, test_offline_dataset.get_ystd()))
        self.assertTrue(np.allclose(train_ymean, train_offline_dataset.get_ymean()))
        self.assertTrue(np.allclose(train_ystd, train_offline_dataset.get_ystd()))


        test_xdim = test_online_dataset._xdata.shape
        self.assertTrue(test_xdim == test_offline_dataset.get_xdim())
        self.assertTrue(test_xdim == test_online_dataset.get_xdim())

if __name__ == "__main__":
    unittest.main()



"""Describes the ModelBaseclass from which other model classes inherit.
"""
import sys
import copy
try:
    import cupy as cp
    from .preconditioners.cuda_rand_nys_preconditioners import Cuda_RandNysPreconditioner
except:
    pass
import numpy as np
from .preconditioners.rand_nys_preconditioners import CPU_RandNysPreconditioner
from .preconditioners.tuning_preconditioners import RandNysTuningPreconditioner
from .preconditioners.rand_nys_constructors import srht_ratio_check

from .kernels import KERNEL_NAME_TO_CLASS
from .constants import constants



class ModelBaseclass():
    """The base class for xGPR model classes. Provides shared
    methods and attributes.

    Attributes:
        kernel_choice (str): The kernel selected by the user.
        kernel: The kernel object for the posterior predictive mean. Set to
            None initially then initialized as soon as the user sets hyperparams /
            runs a tuning routine / fits the model.
        weights: An array, either a cp.ndarray or np.ndarray depending on the
            device specified by the user (cpu or cuda). The random features
            generated by self.kernel are multiplied against the weights to
            generate predictions. The weights are calculated during fitting.
        var: A 2d square array, either a cp.ndarray or np.ndarray depending on
            the device specified by the user (cpu or cuda), or a preconditioner
            object for certain kernels (Linear). The random features
            are used in conjunction with var to generate the posterior predictive
            variance. The var is calculated during fitting. Used by regression
            classes only.
        device (str): One of "cuda", "cpu". The user can update this as desired.
            All predict / tune / fit operations are carried out using the
            current device. If there are multiple GPUs, the one currently
            active is used. You can set the currently active GPU by setting
            an environment variable -- e.g. "export CUDA_VISIBLE_DEVICES=1".
        num_rffs (int): The number of random Fourier features used.
        variance_rffs (int): The number of random Fourier features used for
            calculating posterior predictive variance.
        kernel_settings (dict): Contains kernel-specific parameters --
            e.g. 'matern_nu' for the nu for the Matern kernel, or 'conv_width'
            for the conv1d kernel.
        verbose (bool): If True, regular updates are printed during
            hyperparameter tuning and fitting.
        num_threads (int): The number of threads to use for random feature generation
            if running on CPU. If running on GPU, this argument is ignored.
        double_precision_fht (bool): If True, use double precision during FHT for
            generating random features. For most problems, it is not beneficial
            to set this to True -- it merely increases computational expense
            with negligible benefit -- but this option is useful for testing.
        exact_var_calculation (bool): If True, variance is calculated exactly (within
            the limits of the random feature approximation). If False, a preconditioner
            is used. The preconditioner approach is only used for linear kernels.
        random_seed (int): The seed to the random number generator. Used throughout the
            model class whenever randomness is desired. Can be reset by user.
            Resetting the random seed -- like changing the number of rffs -- will
            cause the kernel to be re-initialized.
        n_classes (int): The number of classes expected in the
            data. This is initialized when fit is called. Used for classification
            only.
        _gamma (ndarray): Either None (if model has not been fitted or is regression)
            or an array of shape (n_classes). Used for classification only.
    """

    def __init__(self, num_rffs:int = 256, variance_rffs:int = 16,
            kernel_choice:str = "RBF", device:str = "cpu",
            kernel_settings:dict = constants.DEFAULT_KERNEL_SPEC_PARMS,
            verbose:bool = True,
            num_threads:int = 2,
            random_seed:int = 123) -> None:
        """Constructor.

        Args:
            num_rffs (int): The number of random Fourier features
                to use. For certain kernels (Linear) this will be set
                by the data.
            variance_rffs (int): The number of random Fourier features
                to use for posterior predictive variance (i.e. calculating
                uncertainty on predictions). Defaults to 64.
            kernel_choice (str): The kernel that the model will use.
                Defaults to 'RBF'. Must be in kernels.kernel_list.
                KERNEL_NAME_TO_CLASS.
            device (str): Determines whether calculations are performed on
                'cpu' or 'cuda'. The initial entry can be changed later
                (i.e. model can be transferred to a different device).
                Defaults to 'cpu'.
            kernel_settings (dict): Contains kernel-specific parameters --
                e.g. 'matern_nu' for the nu for the Matern kernel, or 'conv_width'
                for the conv1d kernel.
            verbose (bool): If True, regular updates are printed
                during fitting and tuning. Defaults to True.
            num_threads (int): The number of threads to use for random feature generation
                if running on CPU. If running on GPU, this argument is ignored.
            random_seed (int): The seed to the random number generator.
        """
        self.kernel_choice = kernel_choice
        self.kernel = None
        self.weights = None
        #Classification classes don't use var or trainy_mean, trainy_std
        self.var = None
        self.trainy_mean = 0
        self.trainy_std = 1

        self.device = device

        self.num_rffs = num_rffs
        #Variance_rffs must be <= num_rffs. Always set second
        self.variance_rffs = variance_rffs

        self.kernel_spec_parms = kernel_settings
        self.num_threads = num_threads

        self.verbose = verbose

        #Currently we do not allow user to set double_precision_fht --
        #only used for testing.
        self.double_precision_fht = False
        # Classification classes don't use exact_var
        self.exact_var_calculation = True
        self.random_seed = random_seed

        #Regression classes don't use n_classes or gamma.
        self.n_classes = 1
        self._gamma = None


    def pre_prediction_checks(self, input_x, sequence_lengths, get_var:bool):
        """Checks input data to ensure validity.

        Args:
            input_x (np.ndarray): A numpy array containing the input data.
            sequence_lengths: None if you are using a fixed-vector kernel (e.g.
                RBF) and a 1d array of the number of elements in each sequence /
                nodes in each graph if you are using a graph or Conv1d kernel.
            get_var (bool): Whether a variance calculation is desired.

        Returns:
            x_array: A cupy array (if self.device is cuda) or a reference
                to the unmodified input array otherwise.

        Raises:
            ValueError: If invalid inputs are supplied,
                a detailed ValueError is raised to explain.
        """
        if self.kernel is None or self.weights is None:
            raise ValueError("Model has not yet been successfully fitted.")
        if not self.kernel.validate_new_datapoints(input_x):
            raise ValueError("The input has incorrect dimensionality.")
        if sequence_lengths is None:
            if len(input_x.shape) != 2:
                raise ValueError("sequence_lengths is required if using a "
                        "convolution kernel.")
        else:
            if len(input_x.shape) == 2:
                raise ValueError("sequence_lengths must be None if using a "
                    "fixed vector kernel.")

        #This should never happen, but just in case.
        if self.weights.shape[0] != self.kernel.get_num_rffs():
            raise ValueError("The size of the weight vector does not "
                    "match the number of random features that are generated.")
        if self.var is None and get_var:
            raise ValueError("Variance was requested but suppress_var "
                    "was selected when fitting, meaning that variance "
                    "has not been generated.")
        if self.device == "cuda":
            mempool = cp.get_default_memory_pool()
            mempool.free_all_blocks()


    def set_hyperparams(self, hyperparams = None, dataset = None):
        """Sets the hyperparameters to those supplied if
        the kernel already exists, creates a kernel if it
        doesn't already exist. If the kernel doesn't already
        exist a dataset needs to be supplied to tell the
        kernel what size its inputs will be. If hyperparams is None,
        the kernel is initialized.

        Args:
            hyperparams (ndarray): A numpy array such that shape[0] == the number of
                hyperparameters for the selected kernel.
            dataset: Either None or a valid dataset object. If the kernel already
                exists (i.e. if set hyperparams or any of the tuning / fitting /
                scoring routines has already been called) no dataset is required
                and this argument is ignored. If the kernel does NOT exist,
                this argument is required.

        Raises:
            ValueError: A ValueError is raised if the kernel does not exist but
                a dataset was not supplied.
        """
        if self.kernel is None and dataset is None:
            raise ValueError("A dataset is required if the kernel has not already "
                    "been initialized. The kernel is initialized by calling set_hyperparams "
                    "or fit or any of the hyperparameter tuning / scoring routines.")
        if hyperparams is None and dataset is None:
            raise ValueError("Should supply hyperparams and/or a dataset.")
        if self.kernel is None:
            self._initialize_kernel(dataset, hyperparams = hyperparams)
        else:
            self.kernel.check_hyperparams(hyperparams)
            self.kernel.set_hyperparams(hyperparams, logspace = True)
            self.weights = None
            self.gamma = None
            self.var = None


    def get_hyperparams(self):
        """Simple helper function to return hyperparameters if the model
        has already been tuned or fitted."""
        if self.kernel is None:
            return None
        return self.kernel.get_hyperparams()



    def _initialize_kernel(self, dataset = None, xdim = None,
            hyperparams = None, bounds = None):
        """Selects and initializes an appropriate kernel object based on the
        kernel_choice string supplied by caller. The kernel is then moved to
        the appropriate device based on the 'device' supplied by caller
        and is returned.

        Args:
            dataset: Either None or a valid dataset object. If None, xdim must not be
                None.
            xdim: Either None or the numpy array returned by dataset.get_xdim(). If None,
                dataset must not be None.
            hyperparams (ndarray): Either None or a numpy array. If not None,
                must be a numpy array such that shape[0] == the number of hyperparameters
                for the selected kernel. The kernel hyperparameters are then initialized
                to the specified value. If None, default hyperparameters are used which
                should then be tuned.
            bounds (np.ndarray): The bounds on hyperparameter
                tuning. Must have an appropriate shape for the
                selected kernel. If None, the kernel will use
                its defaults. Defaults to None.

        Returns:
            kernel: An object of the appropriate kernel class.

        Raises:
            ValueError: Raises a value error if an unrecognized kernel
                is supplied.
        """
        if self.kernel_choice not in KERNEL_NAME_TO_CLASS:
            raise ValueError("An unrecognized kernel choice was supplied.")

        if dataset is None and xdim is not None:
            input_xdim = xdim
        elif dataset is not None:
            input_xdim = dataset.get_xdim()
        else:
            raise ValueError("Either a dataset or xdim must be supplied.")

        self.kernel = KERNEL_NAME_TO_CLASS[self.kernel_choice](input_xdim,
                            self.num_rffs, self.random_seed, self.device,
                            self.num_threads, self.double_precision_fht,
                            kernel_spec_parms = self.kernel_spec_parms)

        #Some kernels set the number of rffs themselves (Linear). If so,
        #reset the number of RFFs for the class to ensure it matches
        #what the kernel is using. We avoid calling the setter and set
        #the attribute directly since the setter may try to reinitialize
        #the kernel under some circumstances.
        self._num_rffs = self.kernel.get_num_rffs()
        if self.variance_rffs >= self.num_rffs:
            raise ValueError("The number of variance rffs must be < num_rffs.")

        if bounds is not None:
            self.kernel.set_bounds(bounds)
        if hyperparams is not None:
            self.kernel.check_hyperparams(hyperparams)
            self.kernel.set_hyperparams(hyperparams, logspace = True)
        self.weights, self.var = None, None


    def _run_pre_nmll_prep(self, dataset, bounds = None):
        """Runs key steps / checks needed if about to calculate
        NMLL.
        """
        if self.kernel is None:
            self._initialize_kernel(dataset, bounds = bounds)
        self.weights, self.var = None, None
        return self.kernel.get_bounds()


    def _run_singlepoint_nmll_prep(self, dataset, exact_method:bool = False):
        """Runs key steps / checks needed if about to calculate
        NMLL at a single point, in which case bounds are not needed.
        """
        if self.kernel is None:
            self._initialize_kernel(dataset)

        self.weights, self.var = None, None

        if self.num_rffs <= 2:
            raise ValueError("num_rffs should be > 2 to use any tuning method.")

        if exact_method:
            if self.kernel.get_num_rffs() > constants.MAX_CLOSED_FORM_RFFS:
                raise ValueError(f"At most {constants.MAX_CLOSED_FORM_RFFS} can be used "
                        "for tuning hyperparameters using this method. Try tuning "
                        "using approximate nmll instead.")



    def _run_pre_fitting_prep(self, dataset, max_rank = None):
        """Runs key steps / checks needed if about to fit the
        model.
        """
        self.trainy_mean = dataset.get_ymean()
        self.trainy_std = dataset.get_ystd()

        if self.kernel is None:
            self._initialize_kernel(dataset)
        if self.variance_rffs > self.kernel.get_num_rffs():
            raise ValueError("The number of variance rffs should be <= the number "
                    "of random features for the kernel.")
        if max_rank is not None:
            if max_rank < 1:
                raise ValueError("Invalid value for max_rank.")
            if max_rank >= self.kernel.get_num_rffs():
                raise ValueError("Max rank should be < the number of rffs.")




    def _check_rank_ratio(self, dataset, sample_frac:float = 0.1,
            max_rank:int = 512):
        """Determines what ratio a particular max_rank would achieve by using a random
        sample of the data. This can be used to determine if a particular max_rank is
        'good enough' to achieve a fast fit, and if so, that max_rank can be used
        for fitting. If the number of rffs is > 8192, this function will use
        8192 by default for better speed (we can get away with this thanks to
        eigenvalue interlacing on random matrices). Note that this procedure
        is not advisable for a small number of datapoints (e.g. < 5,000);
        for those cases, just building the full preconditioner is easier.

        Args:
            dataset: A Dataset object.
            max_rank (int): The maximum rank for the preconditioner, which
                uses a low-rank approximation to the matrix inverse. Larger
                numbers mean a more accurate approximation and thus reduce
                the number of iterations, but make the preconditioner more
                expensive to construct.
            sample_frac (float): The fraction of the data to sample. Must be in
                [0.01, 1].

        Returns:
            achieved_ratio (float): The min eigval of the preconditioner over
                lambda, the noise hyperparameter shared between all kernels.
                This value has decent predictive value for assessing how
                well a preconditioner built with this max_rank is likely to perform.
        """
        if sample_frac < 0.01 or sample_frac > 1:
            raise ValueError("sample_frac must be in the range [0.01, 1]")

        reset_num_rffs = False
        if self.num_rffs > 8192:
            reset_num_rffs, num_rffs = True, copy.deepcopy(self.num_rffs)
            if self.kernel_choice == "RBFLinear" and self.num_rffs % 2 != 0:
                self.num_rffs = 8191
            else:
                self.num_rffs = 8192

        s_mat = srht_ratio_check(dataset, max_rank, self.kernel, self.random_seed,
                self.verbose, sample_frac)
        ratio = float(s_mat.min() / self.kernel.get_lambda()**2) / sample_frac

        if reset_num_rffs:
            self.num_rffs = num_rffs

        return ratio



    def _autoselect_preconditioner(self, dataset, min_rank:int = 512,
            max_rank:int = 3000, increment_size:int = 512,
            always_use_srht2:bool = False,
            ratio_target:float = 30., tuning:bool = False):
        """Uses an automated algorithm to choose a preconditioner that
        is up to max_rank in size. For internal use only.

        Args:
            dataset: A Dataset object.
            max_rank (int): The largest rank that should be used for
                building the preconditioner.
            increment_size (int): The amount by which to increase the
                preconditioner size when increasing the max rank.
            always_use_srht2 (bool): If True, srht2 should always be used
                regardless of the ratio obtained. This will reduce the
                number of iterations about 30% but increase time cost
                of preconditioner construction about 150%.
            ratio_target (int): The target value for the ratio.
            tuning (bool): If True, preconditioner is intended for tuning,
                so it also needs to be used for SLQ.

        Returns:
            preconditioner: A preconditioner object.
        """
        sample_frac, method, ratio, rank = 0.2, "srht", np.inf, min_rank
        actual_num_rffs = self.kernel.get_num_rffs()

        if rank >= actual_num_rffs:
            rank = actual_num_rffs - 1
            ratio = 0.5 * ratio_target

        if dataset.get_ndatapoints() < 5000:
            sample_frac = 1

        while ratio > ratio_target and rank < max_rank:
            ratio = self._check_rank_ratio(dataset, sample_frac = sample_frac,
                    max_rank = rank)
            if ratio > ratio_target:
                if (rank + increment_size) < max_rank and \
                        (rank + increment_size) < actual_num_rffs:
                    rank += increment_size
                else:
                    rank = max_rank
                    if rank > actual_num_rffs:
                        rank = actual_num_rffs - 1
                    method = "srht_2"
                    break

        if self.verbose:
            print(f"Using rank: {rank}")

        if always_use_srht2:
            method = "srht_2"

        if tuning:
            preconditioner = RandNysTuningPreconditioner(self.kernel, dataset, rank,
                        False, self.random_seed, method)
        else:
            if self.device == "cuda":
                preconditioner = Cuda_RandNysPreconditioner(self.kernel, dataset, rank,
                        self.verbose, self.random_seed, method)
                mempool = cp.get_default_memory_pool()
                mempool.free_all_blocks()
            else:
                preconditioner = CPU_RandNysPreconditioner(self.kernel, dataset, rank,
                        self.verbose, self.random_seed, method)

        return preconditioner



    ####The remaining functions are all getters / setters.


    @property
    def kernel_spec_parms(self):
        """Property definition for the kernel_spec_parms."""
        return self._kernel_spec_parms

    @kernel_spec_parms.setter
    def kernel_spec_parms(self, value):
        """Setter for kernel_spec_parms. If the
        user is changing this, the kernel needs to be
        re-initialized."""
        if not isinstance(value, dict):
            raise ValueError("Tried to set kernel_spec_parms to something that "
                    "was not a dict!")
        self._kernel_spec_parms = value
        self.kernel = None
        self.weights = None
        self.gamma = None
        self.var = None


    @property
    def kernel_choice(self):
        """Property definition for the kernel_choice attribute."""
        return self._kernel_choice

    @kernel_choice.setter
    def kernel_choice(self, value):
        """Setter for the kernel_choice attribute. If
        the user is changing this, the kernel needs to
        be re-initialized."""
        if not isinstance(value, str):
            raise ValueError("You supplied a kernel_choice that is not a string.")
        if value not in KERNEL_NAME_TO_CLASS:
            raise ValueError("You supplied an unrecognized kernel.")
        self._kernel_choice = value
        self.kernel = None
        self.weights = None
        self.gamma = None
        self.var = None

    @property
    def num_rffs(self):
        """Property definition for the num_rffs attribute."""
        return self._num_rffs

    @num_rffs.setter
    def num_rffs(self, value):
        """Setter for the num_rffs attribute. If the
        user is changing this, the kernel needs to
        be re-initialized."""
        self._num_rffs = value
        if self.kernel is not None:
            self._initialize_kernel(xdim = self.kernel.get_xdim(),
                   hyperparams = self.kernel.get_hyperparams(),
                   bounds = self.kernel.get_bounds())
        self.weights = None
        self.gamma = None
        self.var = None

    @property
    def variance_rffs(self):
        """Property definition for the variance_rffs attribute."""
        return self._variance_rffs

    @variance_rffs.setter
    def variance_rffs(self, value):
        """Setter for the variance_rffs attribute. If the
        user is changing this, the kernel may need to
        be re-initialized, at least if variance has already
        been calculated. Need to be careful here, because
        for certain fitting procedures, the fit() routine
        does reset variance_rffs."""
        if value > constants.MAX_VARIANCE_RFFS:
            raise ValueError("Currently to keep computational expense at acceptable "
                    f"levels variance rffs is capped at {constants.MAX_VARIANCE_RFFS}.")
        if value > self.num_rffs and self.kernel_choice not in ["Linear", "ExactQuadratic"]:
            raise ValueError("variance_rffs must be <= num_rffs.")
        self._variance_rffs = value
        if self.var is not None:
            if self.kernel is not None:
                self._initialize_kernel(xdim = self.kernel.get_xdim(),
                    hyperparams = self.kernel.get_hyperparams(),
                    bounds = self.kernel.get_bounds())
            self.weights = None
            self.gamma = None
            self.var = None


    @property
    def num_threads(self):
        """Property definition for the num_threads attribute."""
        return self._num_threads

    @num_threads.setter
    def num_threads(self, value):
        """Setter for the num_threads attribute."""
        if value < 1:
            self._num_threads = 1
            raise ValueError("Num threads if supplied must be an integer > 1.")
        self._num_threads = value
        if self.kernel is not None:
            self.kernel.num_threads = value

    @property
    def double_precision_fht(self):
        """Property definition for the double_precision_fht attribute."""
        return self._double_precision_fht


    @double_precision_fht.setter
    def double_precision_fht(self, value):
        """Setter for the double_precision_fht attribute."""
        self._double_precision_fht = value
        if self.kernel is not None:
            self.kernel.double_precision = value

    @property
    def random_seed(self):
        """Property definition for the random_seed attribute."""
        return self._random_seed


    @random_seed.setter
    def random_seed(self, value):
        """Setter for the random_seed attribute. If this is
        reset the kernel needs to be re-initialized."""
        self._random_seed = value
        if self.kernel is not None:
            self._initialize_kernel(xdim = self.kernel.get_xdim(),
                   hyperparams = self.kernel.get_hyperparams(),
                   bounds = self.kernel.get_bounds())
        self.weights = None
        self.gamma = None
        self.var = None


    @property
    def device(self):
        """Property definition for the device attribute."""
        return self._device

    @device.setter
    def device(self, value):
        """Setter for the device attribute."""
        if value not in ["cpu", "cuda"]:
            raise ValueError("Device must be in ['cpu', 'cuda'].")

        if "cupy" not in sys.modules and value == "cuda":
            raise ValueError("You have specified cuda mode but CuPy is "
                "not installed. Currently CPU only fitting is available.")

        if "xGPR.xgpr_cuda_rfgen_cpp_ext" not in sys.modules and value == "cuda":
            raise ValueError("You have specified the cuda fit mode but the "
                "rfgen cuda module is not installed / "
                "does not appear to have installed correctly. "
                "Currently CPU only fitting is available.")

        if self.kernel is not None:
            self.kernel.device = value
        if self.weights is not None:
            if value == "cuda":
                self.weights = cp.asarray(self.weights)
            elif value == "cpu" and not isinstance(self.weights, np.ndarray):
                self.weights = cp.asnumpy(self.weights)
        if self.var is not None:
            if self.exact_var_calculation:
                if value == "cpu":
                    self.var = cp.asnumpy(self.var)
                else:
                    self.var = cp.asarray(self.var)
            else:
                self.var.device = value
        if value == "cuda":
            mempool = cp.get_default_memory_pool()
            mempool.free_all_blocks()
        self._device = value



"""Tests the auxiliary tools (kernel pca, rf gen for kernel kmeans)
for obvious failures. This doesn't do any performance tests since
both are just wrappers on functionality used elsewhere; it
just ensures that they are functional."""
import sys
import os
import unittest

import numpy as np

from xGPR import KernelFGen

#TODO: Get rid of this path alteration
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset
from utils.model_constructor import get_models
from utils.evaluate_model import evaluate_model


class CheckAuxiliaryFunctions(unittest.TestCase):
    """Checks whether the feature gen for
    kernel k-means is functioning."""

    def test_kernel_fgen(self):
        """Test the kernel fgen functionality."""
        online_sdata, _ = build_test_dataset(conv_kernel = False)
        online_cdata, _ = build_test_dataset(conv_kernel = True)

        #Ensure no exception if building the tool for standard datasets with standard
        #kernels and doing the usual things.
        raised = False
        kpca = KernelFGen(num_rffs = 500,
                            kernel_choice = "RBF", hyperparams = np.array([1.0]),
                            num_features = online_sdata.get_xdim()[-1],
                            random_seed = 123)
        xtrans = kpca.predict(online_sdata._xdata)
        self.assertFalse(raised, 'Does not work with RBF kernel.')
        self.assertTrue(xtrans.shape[1] == 500)

        raised = False
        try:
            kpca = KernelFGen(num_rffs = 500,
                            kernel_choice = "GraphRBF", hyperparams = np.array([1.0]),
                            num_features = online_cdata.get_xdim()[-1],
                            random_seed = 123)
            xtrans = kpca.predict(online_cdata._xdata, online_cdata._sequence_lengths)
        except:
            raised = True
        self.assertFalse(raised, 'Does not work with graph kernel.')
        self.assertTrue(xtrans.shape[1] == 500)

if __name__ == "__main__":
    unittest.main()



"""Describes the Conv1dKernelBaseclass, which is shared
as a parent class by several convolution kernels for
graphs and sequences."""
from abc import ABC
from math import ceil

import numpy as np
from scipy.stats import chi

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuConv1dFGen, cpuConvGrad
try:
    import cupy as cp
    from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaConv1dFGen, cudaConvGrad
except:
    pass

from ..kernel_baseclass import KernelBaseclass



class ConvKernelBaseclass(KernelBaseclass, ABC):
    """The baseclass for structured orthogonal random features (SORF)
    kernels that do 1d convolution. Since it inherits from
    KernelBaseclass, it includes the attributes of that class. Only
    additional attributes unique to this class are described here.

    Attributes:
        conv_width (int): The width of the convolution kernel.
            This hyperparameter can be set based on experimentation
            or domain knowledge.
        radem_diag: The diagonal matrices for the SORF transform. Type is int8.
        chi_arr: A diagonal array whose elements are drawn from the chi
            distribution. Ensures the marginals of the matrix resulting
            from S H D1 H D2 H D3 are correct.
        sequence_average (bool): If True, the features are averaged over the sequence
            when summing. Can be set to True by passing "averaging":True under
            kernel_spec_parms, otherwise defaults to False. This is useful if
            modeling properties of a sequence that are not size-extensive.
    """

    def __init__(self, xdim, num_rffs, random_seed = 123,
                    num_threads = 2, double_precision = False,
                    conv_width = 9, kernel_spec_parms = {}):
        """Constructor.

        Args:
            xdim (tuple): The dimensions of the input. Either (N, D) or (N, M, D)
                where N is the number of datapoints, D is number of features
                and M is number of timepoints or sequence elements (convolution
                kernels only).
            num_rffs (int): The user-requested number of random Fourier features.
                For sine-cosine kernels (RBF, Matern), this will be saved by the
                class as num_rffs.
            random_seed (int): The seed to the random number generator.
            device (str): One of 'cpu', 'cuda'. Indicates the starting device.
            num_threads (int): The number of threads to use for random feature generation
                if running on CPU. If running on GPU, this is ignored.
            double_precision (bool): If True, generate random features in double precision.
                Otherwise, generate as single precision.
            conv_width (int): The width of the convolution to perform.
            kernel_spec_parms (dict): A dictionary of additional kernel-specific
                attributes.

        Raises:
            RuntimeError: A RuntimeError is raised if the dimensions of the input are
                inappropriate given the conv_width.
        """
        super().__init__(num_rffs, xdim, num_threads = 2,
                sine_cosine_kernel = True, double_precision = double_precision,
                kernel_spec_parms = kernel_spec_parms)
        if len(xdim) != 3:
            raise RuntimeError("Tried to initialize a Conv1d kernel with a 2d x-"
                    "array! x should be a 3d array for Conv1d.")

        self.scaling_type = 0
        if "averaging" in kernel_spec_parms:
            if kernel_spec_parms["averaging"] == "none":
                self.scaling_type = 0
            elif kernel_spec_parms["averaging"] == "sqrt":
                self.scaling_type = 1
            elif kernel_spec_parms["averaging"] == "full":
                self.scaling_type = 2
            else:
                raise RuntimeError("Unrecognized value for 'averaging' supplied, "
                        "should be one of 'none', 'sqrt', 'full'.")

        rng = np.random.default_rng(random_seed)
        self.conv_width = conv_width

        dim2_no_padding = self.conv_width * xdim[2]
        padded_dims = 2**ceil(np.log2(max(dim2_no_padding, 2)))
        init_calc_freqsize = ceil(self.num_freqs / padded_dims) * \
                        padded_dims

        radem_array = np.asarray([-1,1], dtype=np.int8)
        self.radem_diag = rng.choice(radem_array, size=(3, 1, init_calc_freqsize),
                                replace=True)
        self.chi_arr = chi.rvs(df=padded_dims, size=self.num_freqs,
                            random_state = random_seed)
        if not self.double_precision:
            self.chi_arr = self.chi_arr.astype(np.float32)




    def kernel_specific_set_device(self, new_device):
        """Called by parent class when device is changed. Moves
        some of the object parameters to the appropriate device."""
        if new_device == "cuda":
            self.radem_diag = cp.asarray(self.radem_diag)
            self.chi_arr = cp.asarray(self.chi_arr)
        elif not isinstance(self.radem_diag, np.ndarray):
            self.radem_diag = cp.asnumpy(self.radem_diag)
            self.chi_arr = cp.asnumpy(self.chi_arr)



    def kernel_specific_transform(self, input_x, sequence_length):
        """Generates random features.

        Args:
            input_x: A numpy or cupy array containing the raw input data.

        Returns:
            xtrans: A cupy or numpy array containing the generated features.
            sequence_length: A numpy or cupy array containing the number of
                elements in each sequence -- so that zero padding can be masked.

        Raises:
            RuntimeError: A value error is raised if the dimensionality of the
                input does not meet validity criteria.
        """
        if sequence_length is None:
            raise RuntimeError("sequence_length is required for convolution kernels.")
        if input_x.shape[2] != self._xdim[2]:
            raise RuntimeError("Unexpected input shape supplied.")

        input_x *= self.hyperparams[1]

        if self.device == "cpu":
            xtrans = np.zeros((input_x.shape[0], self.num_rffs), np.float64)
            cpuConv1dFGen(input_x, xtrans, self.radem_diag, self.chi_arr,
                    sequence_length, self.conv_width, self.scaling_type,
                    self.num_threads)
        else:
            xtrans = cp.zeros((input_x.shape[0], self.num_rffs), cp.float64)
            cudaConv1dFGen(input_x, xtrans, self.radem_diag, self.chi_arr,
                    sequence_length, self.conv_width, self.scaling_type)

        return xtrans


    def kernel_specific_set_hyperparams(self):
        """Provided for consistency with baseclass. This
        kernel has no kernel-specific properties that must
        be reset after hyperparameters are changed."""
        return


    def kernel_specific_gradient(self, input_x, sequence_length):
        """The gradient for kernel-specific hyperparameters is calculated
        using an array (dz_dsigma) specific to each kernel.

        Args:
            input_x: A cupy or numpy array containing the raw input data.
            sequence_length: A numpy or cupy array containing the number of
                elements in each sequence -- so that zero padding can be masked.

        Returns:
            output_x: A cupy or numpy array containing the random feature
                representation of the input.
            dz_dsigma: A cupy or numpy array containing the derivative of
                output_x with respect to the kernel-specific hyperparameters.
        """
        if sequence_length is None:
            raise RuntimeError("sequence_length is required for convolution kernels.")
        if input_x.shape[2] != self._xdim[2]:
            raise RuntimeError("Unexpected input shape supplied.")

        if self.device == "cpu":
            xtrans = np.zeros((input_x.shape[0], self.num_rffs), np.float64)
            dz_dsigma = np.zeros((input_x.shape[0], self.num_rffs, 1), np.float64)
            cpuConvGrad(input_x, xtrans, self.radem_diag, self.chi_arr,
                    sequence_length, dz_dsigma, self.hyperparams[1],
                    self.conv_width, self.scaling_type,
                    self.num_threads)
        else:
            xtrans = cp.zeros((input_x.shape[0], self.num_rffs), cp.float64)
            dz_dsigma = cp.zeros((input_x.shape[0], self.num_rffs, 1), cp.float64)
            cudaConvGrad(input_x, xtrans, self.radem_diag, self.chi_arr,
                    sequence_length, dz_dsigma, self.hyperparams[1],
                    self.conv_width, self.scaling_type)

        return xtrans, dz_dsigma



"""Provides raw data loading services for other unit tests."""
import os

import numpy as np

from xGPR.data_handling.dataset_builder import build_regression_dataset

RANDOM_STATE = 123


def build_test_dataset(conv_kernel = False, xsuffix = "trainxvalues.npy",
        ysuffix = "trainyvalues.npy"):
    """Loads the test data provided with xGPR and converts it
    into an OfflineDataset object and an OnlineDataset object.
    Both are returned. Used by multiple other unit-tests.

    Args:
        conv_kernel (bool): If True, return the 3d array data
            for the convolution kernel. If False, return
            the 2d data for basic kernels.
        xsuffix (str): The expected ending for the file list that
            will be retrieved for x-data.
        ysuffix (str): The expected ending for the file list that
            will be retrieved for y-data.

    Returns:
        online_data (OnlineDataset): The raw data stored in memory.
        offline_data (OfflineDataset): The raw data stored on disk.
    """
    start_path = os.path.abspath(os.path.dirname(__file__))
    if not conv_kernel:
        os.chdir(os.path.join(start_path, "..", "test_data"))
        sequence_lengths = None
    else:
        os.chdir(os.path.join(start_path, "..", "test_data", "conv_test"))

    xtrain_files = [f for f in os.listdir() if f.endswith(xsuffix)]
    xtrain_files.sort()
    ytrain_files = [f.replace(xsuffix, ysuffix) for f in
            xtrain_files]
    if conv_kernel:
        sequence_lengths = [f.replace("xvalues.npy", "seqlen.npy")
                for f in xtrain_files]

    offline_data = build_regression_dataset(xtrain_files,
                ytrain_files, sequence_lengths, chunk_size = 2000)

    if conv_kernel:
        xvalues, yvalues, seqlen = [], [], []
        for xfile, yfile, lfile in zip(xtrain_files, ytrain_files, sequence_lengths):
            xvalues.append(np.load(xfile))
            yvalues.append(np.load(yfile))
            seqlen.append(np.load(lfile))
        seqlen = np.concatenate(seqlen)
    else:
        xvalues, yvalues, seqlen = [], [], None
        for xfile, yfile in zip(xtrain_files, ytrain_files):
            xvalues.append(np.load(xfile))
            yvalues.append(np.load(yfile))


    xvalues = np.vstack(xvalues)
    yvalues = np.concatenate(yvalues)
    online_data = build_regression_dataset(xvalues, yvalues, seqlen, chunk_size = 2000)

    return online_data, offline_data



def build_traintest_split(conv_kernel = False, xsuffix = "trainxvalues.npy",
        ysuffix = "trainyvalues.npy"):
    """Loads the test data provided with xGPR and converts it
    into a train dataset and a test dataset.

    Args:
        conv_kernel (bool): If True, return the 3d array data
            for the convolution kernel. If False, return
            the 2d data for basic kernels.
        xsuffix (str): The expected ending for the file list that
            will be retrieved for x-data.
        ysuffix (str): The expected ending for the file list that
            will be retrieved for y-data.

    Returns:
        train_data (OnlineDataset): The training data.
        test_data (OnlineDataset): The test data.
    """
    _, online_data = build_test_dataset(conv_kernel, xsuffix, ysuffix)

    xvalues = online_data.get_xdata()
    yvalues = online_data.get_ydata()
    seqlen = online_data.get_sequence_lengths()

    rng = np.random.default_rng(123)
    idx = rng.permutation(xvalues.shape[0])
    xvalues, yvalues = xvalues[idx,:], yvalues[idx]
    cutoff = int(0.75 * idx.shape[0])

    if seqlen is None:
        train_data = build_regression_dataset(xvalues[:cutoff,...], yvalues[:cutoff],
                chunk_size = 2000)
        test_data = build_regression_dataset(xvalues[cutoff:,...], yvalues[cutoff:],
                chunk_size = 2000)
    else:
        train_data = build_regression_dataset(xvalues[:cutoff,...], yvalues[:cutoff],
                seqlen[:cutoff], chunk_size = 2000)
        test_data = build_regression_dataset(xvalues[cutoff:,...], yvalues[cutoff:],
                seqlen[cutoff:], chunk_size = 2000)

    return train_data, test_data
